// Copyright 1996-2011 by Jon Dart.  All Rights Reserved.

#include "globals.h"
#include "hash.h"
#include "bitprobe.h"
#include "scoring.h"
#include "tbprobe.h"
#include "bitbase.cpp"

char cmd_buf[CMD_BUF_SIZE];
int cmd_buf_count;
int EGTBMenCount = 0;
MoveArray *gameMoves;
Options options;
Book *openingBook = NULL;
Log *theLog = NULL;
string learnFileName;
#ifdef _WIN32
HANDLE inputEvent;
#endif
LockDefine(input_lock);

#define LEARN_FILE_NAME "arasan.lrn"

string programPath;

string derivePath(const char *fileName) {
   return derivePath(programPath.c_str(),fileName);
}

string derivePath(const char *base, const char *fileName) {
    string result(base);
    size_t pos;
#ifdef _WIN32
    pos = result.rfind('\\',string::npos);
#else
    pos = result.rfind('/',string::npos);
#endif
    if (pos == string::npos) {
      return fileName;
    }
    else {
      return result.substr(0,pos+1) + fileName;
    }
}

int initGlobals(const char *pathName) {
   programPath = pathName;
   gameMoves = new MoveArray();
   theLog = new Log();
   theLog->clear();
   theLog->write_header();
#ifdef _WIN32
   inputEvent = CreateEvent(NULL,FALSE,FALSE,NULL);
   LockInit(input_lock);
#endif
   return 1;
}

void CDECL cleanupGlobals(void) {
   delete openingBook;
   delete gameMoves;
   delete theLog;
   LockDestroy(input_lock);
   Scoring::cleanup();
#ifdef _WIN32
   CloseHandle(inputEvent);
#endif
}

void initOptions(const char *pathName) {
    programPath = pathName;
    string rcPath = derivePath("arasan.rc");
    // try to read arasan.rc file
    options.init(rcPath);
#ifndef _WIN32
    // Also read .rc from the user's HOME,
    // if there is one.
    if (getenv("HOME")) {
       rcPath = derivePath(getenv("HOME"),"arasan.rc");
       options.init(rcPath);       
    }
#endif
    learnFileName = derivePath(LEARN_FILE_NAME);
}

void delayedInit() {
    // If no tablebase directory was specified in the .rc file,
    // try to find the table bases under the arasan install dir
    if (options.search.use_tablebases &&
        options.search.tablebase_location == "") {
       options.search.tablebase_location=derivePath("TB");
    }
    if (options.search.use_tablebases) {
       EGTBMenCount = initTB((char*)options.search.tablebase_location.c_str(),
         options.search.tb_cache_size);
    }
    // also initialize the book here
    if (options.book.book_enabled) {
       openingBook = new Book();
    }
}
